/*
-- ****************************************************************
-- ***  Creacin de Tablas Particionadas en SQL Server,         ***
-- ***  y movimiento de Particiones entre Tablas Particionadas  ***
-- ****************************************************************
--
-- GuilleSQL - Un portal sobre Microsoft SQL Server en Castellano
-- http://www.guillesql.es
--
*/

/*
Creamos una Funcin de Particionamiento, para crear 8 particiones:
Particin 1: Valores < 20100101 (Diciembre 2009 y meses anteriores)
Particin 2: 20100101 <= Valores < 20100201 (Enero 2010)
Particin 3: 20100201 <= Valores < 20100301 (Febrero 2010)
Particin 4: 20100301 <= Valores < 20100401 (Marzo 2010)
Particin 5: 20100401 <= Valores < 20100501 (Abril 2010)
Particin 6: 20100501 <= Valores < 20100601 (Mayo 2010)
Particin 7: 20100601 <= Valores < 20100701 (Junio 2010)
Particin 8: 20100701 <= Valores (Julio 2010 y meses posteriores)
Al trabajar con fechas, utilizaremos RANGE RIGHT (suele ser lo habitual)
NOTA: Suponemos que las particin 1 y 8 estn vacas, y el resto contienen datos.
La idea, es mantener inicialmente informacin de los meses de Enero a Junio.
*/

CREATE PARTITION FUNCTION [pfDatos] (datetime)
AS RANGE RIGHT FOR VALUES 
('20100101', '20100201', '20100301', '20100401', '20100501', '20100601', '20100701');
GO

/*
Creamos un Esquema de Particionamiento, 
para mapear todas las particiones a un mismo FileGroup
*/

CREATE PARTITION SCHEME [psDatos] 
AS PARTITION [pfDatos] ALL TO ([PRIMARY]);
GO

/*
Creamos una Funcin de Particionamiento, para crear 2 particiones:
Particin 1: Valores < 20100101 (Diciembre 2009 y meses anteriores)
Particin 2: 20100101 <= Valores (Enero 2010 y meses posteriores)
El objetivo es mantener la Particin 1 llena, y la particin 2 vaca
De este modo, la particin 2, al estar vaca permitir realizar SPLIT de forma eficiente.
Adems, despus del SPLIT, la nueva particin, podr recibir datos con ALTER TABLE SWITCH PARTITION,
para seguidamente poder hacer un MERGE, y volver a la misma situacin inicial:
Dos particiones, con la Particin 1 llena, y la particin 2 vaca.
*/

CREATE PARTITION FUNCTION [pfDatosHistoricos] (datetime)
AS RANGE RIGHT FOR VALUES 
('20100101');
GO

/*
Creamos un Esquema de Particionamiento, 
para mapear todas las particiones a un mismo FileGroup
*/

CREATE PARTITION SCHEME [psDatosHistoricos] 
AS PARTITION [pfDatosHistoricos] ALL TO ([PRIMARY]);
GO


/*
Creamos una tabla con varias particiones
para los datos recientes y frecuentemente accedidos
*/

CREATE TABLE dbo.tblDatos
(
ID UNIQUEIDENTIFIER NOT NULL
,DESCRIPCION VARCHAR(200) NOT NULL
,FECHA DATETIME NOT NULL
,CONSTRAINT	PK_tblDatos PRIMARY KEY NONCLUSTERED (ID, FECHA) ON psDatos(FECHA)
) ON psDatos(FECHA);

/*
Creamos una tabla con dos particiones
para los datos histricos y poco accedidos
Debe tener el mismo esquema que la tabla anterior, para poder hacer el SWITCH !!
*/

CREATE TABLE dbo.tblDatosHistoricos
(
ID UNIQUEIDENTIFIER NOT NULL
,DESCRIPCION VARCHAR(200) NOT NULL
,FECHA DATETIME NOT NULL
,CONSTRAINT	PK_tblDatosHistoricos PRIMARY KEY NONCLUSTERED (ID, FECHA) ON psDatosHistoricos(FECHA)
) ON psDatosHistoricos(FECHA);

/*
Insertamos datos de prueba en la tabla de datos recientes, y los consultamos
*/

INSERT INTO dbo.tblDatos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Junio 2010','20100601')
INSERT INTO dbo.tblDatos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Mayo 2010','20100501')
INSERT INTO dbo.tblDatos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Abril 2010','20100401')
INSERT INTO dbo.tblDatos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Marzo 2010','20100301')
INSERT INTO dbo.tblDatos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Febrero 2010','20100201')
INSERT INTO dbo.tblDatos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Enero 2010','20100101')

SELECT *, $PARTITION.pfDatos(FECHA) [Partition] FROM dbo.tblDatos 

/*
Insertamos datos de prueba en la tabla de datos histricos, y los consultamos
*/

INSERT INTO dbo.tblDatosHistoricos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Dicembre 2009','20091201')
INSERT INTO dbo.tblDatosHistoricos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Noviembre 2009','20091101')
INSERT INTO dbo.tblDatosHistoricos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Octubre 2009','20091001')
INSERT INTO dbo.tblDatosHistoricos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Septiembre 2009','20090901')
INSERT INTO dbo.tblDatosHistoricos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Agosto 2009','20090801')
INSERT INTO dbo.tblDatosHistoricos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Julio 2009','20090701')

SELECT *, $PARTITION.pfDatosHistoricos(FECHA) [Partition] FROM dbo.tblDatosHistoricos 

/*
Ahora llega el momento de gracia.
Evidentemente, antes o despus, necesitaremos pasar el ltimo mes de tblDatos (Enero 2010) a tblDatosHistoricos,
y adems en tblDatos se empezarn a recibir datos para el nuevo mes (Julio 2010).
Por ello, debemos tanto archivar el mes de Enero 2010 en la tabla histrica,
como acondicionar tblDatos para recibir datos de Julio 2010.
El primer paso, hacer un SPLIT en tblDatos, para crear una particin para el mes de Julio 2010.
Por qu? Porque en este momento, la particin sobre la que vamos a hacer el SPLIT est vaca,
por lo tanto, el SPLIT ser una operacin eficiente (metadata only operation).
Por ello, deberemos hacerlo ahora, y no dejar que avance el tiempo. Vamos a ello.
*/

ALTER PARTITION FUNCTION [pfDatos] () SPLIT RANGE ('20100801');
-- Ojo: Deberamos ejecutar la siguiente lnea, pero no lo vamos a hacer
-- para que luego casque, y as veamos claramente su necesidad. Prueba y error !
-- ALTER PARTITION SCHEME [psDatos] NEXT USED [PRIMARY]

/*
Realizado esto, ya se podran insertar datos de Julio 2010,
existiendo una particin adicional para Agosto 2010 y meses posteriores,
la cual se mantendr vaca.
As, podremos repetir el proceso todos los meses, es decir, hacer SPLIT de una particin vaca.
Bien. Ahora vamos a continuar con el archivado (SWITCH) del mes de Enero 2010.
Por ello, el siguiente paso es hacer un SPLIT sobre pfDatosHistoricos.
Por qu? Porque en este momento, la particin de destino para el SWITCH est vaca,
por lo que podemos hacer un SPLIT, aprovechando la mejora de rendimiento de esta situacin,
para seguidamente realizar el SWITCH.
As, podremos repetir el proceso todos los meses, es decir, hacer SPLIT de una particin vaca.
*/

ALTER PARTITION FUNCTION [pfDatosHistoricos] () SPLIT RANGE ('20100201');
-- Ojo: Deberamos ejecutar la siguiente lnea, pero no lo vamos a hacer
-- para que luego casque, y as veamos claramente su necesidad. Prueba y error !
-- ALTER PARTITION SCHEME [psDatosHistoricos] NEXT USED [PRIMARY]

/*
Genial. Hecho esto, continuamos.
Para poder mover una particin (archivar), sta debe ser un subconjunto de la particin destino.
Queremos mover la particin 2 de dbo.tblDatos, de rango:  20100101 <= Valores < 20100201 (Enero 2010)
utilizando como destino la particin 2 de dbo.tblDatosHistoricos, de rango:  20100101 <= Valores < 20100201 (Enero 2010)
Tanto las tablas como los ndices, tienen que estar particionados
La particin destino debe estar vaca
Las particiones origen y destino deben estar en el mismo FileGroup
Esta operacin es una operacin de metadatos (metadata only operation),
por lo que resulta muy eficiente (rpida y sin crecimiento de Log).
*/

ALTER TABLE dbo.tblDatos SWITCH PARTITION 2 TO dbo.tblDatosHistoricos PARTITION 2

/*
Ahora, si comprobamos los datos, hemos conseguido mover los datos de Enero
desde la tabla origen a la tabla destino.
Claro que en dbo.tblDatos, sigue existiendo una particin 2 para Enero 2010,
aunque eso s, despus del SWITCH, dicha particin est vaca.
Adems en dbo.tblDatosHistoricos tenemos tres particiones, 2 con datos y una vaca.
An quedan cosas por hacer.
*/

SELECT *, $PARTITION.pfDatos(FECHA) [Partition] FROM dbo.tblDatos
SELECT *, $PARTITION.pfDatosHistoricos(FECHA) [Partition] FROM dbo.tblDatosHistoricos 

/*
Ahora toca mezclar (MERGE) el mes de Enero con el resto de meses en la tabla histrica
NOTA: esta operacin es costosa, ya que las particiones a mezclar no estn vacas:
puede incurrir en un elevado tiempo de ejecucin y crecimiento de Log !!!
Una alternativa, es mantener en la tabla Histrica particiones para cada ao, 
en vez de una nica particin para todo, o incluso mantener una particin para cada mes evitando el MERGE.
Como siempre, es una cuestin de diseo, que se debe examinar con calma en cada caso.
*/

ALTER PARTITION FUNCTION [pfDatosHistoricos] () MERGE RANGE ('20100101');

/*
Ahora toca alterar la funcin de particionamiento correspondiente a tblDatos,
para mezclar la particin 2 de Enero 2010 (que est vaca despus del SPLIT)
con la particin 1 de Diciembre 2010 y meses posteriores (tambin vaca).
Al estar vacas ambas particiones, se tratara de una operacin de metadatos,
con la correspondiente mejora de rendimiento.
*/

ALTER PARTITION FUNCTION [pfDatos] () MERGE RANGE ('20100101');

/*
Ahora podemos insertar datos para el nuevo mes de Julio 2010, 
que se insertarn en la particin correspondiente,
dejando la ltima particin vaca, para cuando sea necesario hacer otro SPLIT
*/

INSERT INTO dbo.tblDatos (ID, DESCRIPCION, FECHA) VALUES (NEWID(), 'Julio 2010','20100701')

/*
As queda el contenido de las tablas y la pertenencia de filas a particiones
En tblDatos, continan las particiones 1 y 8 vacas, y el resto llenas.
En tblDatosHistoricos, contina con una particin llena, y otra vaca.
Por lo tanto, el proceso es repetible. 
Dentro de un mes, podremos repetirlo de igual forma, 
quedando una configuracin equivalente de particiones llenas y vacas, etc.
*/

SELECT *, $PARTITION.pfDatos(FECHA) [Partition] FROM dbo.tblDatos
SELECT *, $PARTITION.pfDatosHistoricos(FECHA) [Partition] FROM dbo.tblDatosHistoricos 

/*
Si tenemos curiosidad, podemos consultar la configuracin de particiones,
segn las tablas/vistas del sistema. Por ejemplo:
*/

SELECT FN.name, FN.fanout, VL.boundary_id, VL.value 
FROM sys.partition_functions FN
INNER JOIN sys.partition_range_values VL
ON FN.function_id = VL.function_id

/*
Hemos acabado ya? Pues no. Hay un detalle ms que es interesante comentar.
En teora, el proceso que hemos llevado a cabo es repetible Verdad?
La configuracin de particiones ahora mismo, es equivalente a como era al principio.
Por lo tanto, deberamos poder hacer nuevos SPLITs Verdad? Pues no. Vase:
*/

ALTER PARTITION FUNCTION [pfDatos] () SPLIT RANGE ('20100901');

/*
Cspita !!! Parece que nos insulta. 
Al crear las funciones y esquemas de particionamientos de nuestro ejemplo, 
se configura de manera implcita el FileGroup PRIMARY como siguiente FileGroup (NEXT USED)
a ser utilizado en una operacin SPLIT.
Por ello, al hacer un primer SPLIT todo funciona OK, 
pero para hacer un siguiente SPLIT debemos indicarle a SQL Server cual va a ser el NEXT USED FileGroup,
y sino lo hacemos, pues nada: Error al canto !!! 
Conclusin: ahora para salir del apuro:
*/

ALTER PARTITION SCHEME [psDatos] NEXT USED [PRIMARY]
ALTER PARTITION FUNCTION [pfDatos] () SPLIT RANGE ('20100901');
ALTER PARTITION SCHEME [psDatos] NEXT USED [PRIMARY]

/*
Moraleja, debemos tener como costubre en estos casos, despus de cada SPLIT,
establecer el FileGroup deseado como NEXT USED, para as evitarnos sustos.
Pero an queda otro punto que ver: 
Qu ocurre si intentamos hacer un SWITCH, SPLIT o MERGE con actividad de usuarios sobre las particiones?
Este tipo de operaciones, generan bloqueos de esquema (locks), 
por lo podemos encontrarnos con esperas (waits) debidas a bloqueos (blocks).
Esto se puede ver iniciando en una sesin una transaccin, pero sin finalizarla:
*/


BEGIN TRAN

UPDATE dbo.tblDatos SET DESCRIPCION=DESCRIPCION WHERE DESCRIPCION='Febrero 2010'


/*
En otra sesin, intentamos hacer un SWITCH
*/

ALTER PARTITION FUNCTION [pfDatos] () SPLIT RANGE ('20101001');
ALTER PARTITION SCHEME [psDatos] NEXT USED [PRIMARY]

/*
Vaya !! el SWITCH no progresa, y se queda en espera de que se libere el bloqueo existente.
Hacemos un COMMIT en la transaccin anterior:
*/

COMMIT

/*
En este momento, se ha finalizado la transaccin, y adems, se ha ejecutado por fin el SPLIT 
que estaba esperando por el bloqueo (lock/block).
Y ahora, s que hemos terminado !!!
*/